/* 
  AgentManager.m

  Barry McMullin <mcmullin@eeng.dcu.ie>
  07-OCT-1996

*/

#import <collections.h>
#import <random.h>

#import "SCLGlobals.h"
#import "Agent.h"
#import "AgentManager.h"

@implementation AgentManager

+createBegin: (id) aZone  {
  AgentManager * mgr;

  mgr = [super createBegin: aZone];

  mgr->agentList = [List create: aZone];
  mgr->addedAgentQueue = [List create: aZone];
  mgr->droppedAgentQueue = [List create: aZone];

  return mgr;
}

-setAgentClass: (id) classId {
  if (agentClass != nil)
    [InternalError raiseEvent:
      "AgentManager %d: attempt to setAgentClass when already set to %d.\n",
      self, agentClass];
  agentClass = classId;

  return self;
}

-createEnd {
  if (agentClass == nil)
    [InternalError raiseEvent:
      "AgentManager object %d not fully initialised before createEnd.\n",
      self];

  return [super createEnd];
}

-(void) setUniformUnsigned: (id) inUniformUnsigned {
  uniformUnsigned = inUniformUnsigned;
}

-(id) getUniformUnsigned {
  return uniformUnsigned;
}


-(id) getAgentList {
  return agentList;
}

-(id) createAgent {
  id agent;

  agent = [agentClass create: [self getZone]];
  [addedAgentQueue addLast: agent];

  return agent;
}


-dropAgent: (id) agent {
  if ([droppedAgentQueue contains: agent])
    [InternalError raiseEvent: 
      "Attempt to drop already-dropped agent (%p)\n", agent];
  else
    [droppedAgentQueue addLast: agent];

  return self;
}

-(unsigned) getCount {
  return [agentList getCount];
}


-(id) randomAgent {
  int offset;
  int count;
  id agent;

  agent = nil;
  count = [agentList getCount];
  if (count > 0) {
    offset = [uniformUnsigned getUnsignedWithMin:0 withMax: (count - 1)];
    agent = [agentList atOffset: offset];
  }

  return agent;
}


-processQueues {
  id agent;

  while ([addedAgentQueue getCount] > 0) {
    agent = [addedAgentQueue getFirst];
    [agentList addLast: agent];
    [addedAgentQueue removeFirst];
  }

  while ([droppedAgentQueue getCount] > 0) {
    agent = [droppedAgentQueue getFirst];
    [agentList remove: agent];
    [droppedAgentQueue removeFirst];
    [agent drop];
  }

  return self;
}

-saveTo: (id <OutFile>) file {
  int i, count;

  [self processQueues];  // Defensive...

  count = [agentList getCount];
  [file putString: "# Start agentList (Class: "];
  [file putString: [agentClass name]];
  [file putString: ")\n"];
  [file putString: "# Agent count:\n"];
  [file putInt: count];
  [file putString: "\n# Agents:\n"];

  for (i = 0; i < count; i++) {
    [[agentList atOffset: i] saveTo: file];
    [file putNewLine];
  }

  return self;
}

-loadFrom: (id <InFile>) file toLattice: (id) inLattice {
  int i, count;
  id agent;

  [self processQueues];  // Defensive...

  [file skipLine]; // "# Start agentList (Class: ... )\n"
  [file skipLine]; // "# Agent count:\n"
  [file getInt: &count];
  [file skipLine]; 
  [file skipLine]; // "# Agents:\n"];

  for (i = 0; i < count; i++) {
    agent = [self createAgent];
    [agent loadFrom: file toLattice: inLattice];
    [file skipLine];
  }

  [self processQueues];

  return self;
}



-(void) drop {
  id agent;

  [self processQueues]; // Defensive...

  while ([agentList getCount] > 0) {
    agent = [agentList getFirst];
    [agentList remove: agent];
    [agent drop];
  }

  [agentList drop];
  [addedAgentQueue drop];
  [droppedAgentQueue drop];

  [super drop];
}

-(void) xprintStatus {
  printf("AgentManager: xprintStatus (self: %p)\n", self);
  printf("  agentClass: %p\n", agentClass);
  printf("  agentList (count = %d):\n", [agentList getCount]);
  fflush(stdout);
  [agentList forEach: M(xprintStatus)];
  printf("  addedAgentQueue (count = %d):\n", [addedAgentQueue getCount]);
  fflush(stdout);
  [addedAgentQueue forEach: M(xprintStatus)];
  printf("  droppedAgentQueue (count = %d):\n", [droppedAgentQueue getCount]);
  fflush(stdout);
  [droppedAgentQueue forEach: M(xprintStatus)];
}


@end
