/* 
  Catalyst.m

  Barry McMullin <mcmullin@eeng.dcu.ie>
  30-SEP-1996

*/

#import <simtools.h>

#import "SCLGlobals.h"
#import "SCLModelSwarm.h"

#import "Bernoulli.h"
#import "Lattice.h"
#import "ParticleManager.h"
#import "Hole.h"
#import "Substrate.h"
#import "Link.h"
#import "Catalyst.h"

@implementation Catalyst


static id bernoulli;
static double catalystMobilityFactor = 0.0;

+(void) setBernoulli: (id) inBernoulli {
  bernoulli = inBernoulli;
}

+(id) getBernoulli {
  return(bernoulli);
}

+(void) setClassMobilityFactor: (double) inMobilityFactor {
  catalystMobilityFactor = inMobilityFactor;
}

+(double) getClassMobilityFactor {
  return(catalystMobilityFactor);
}


static BOOL catSmartMotionFlag;

+(void) setCatSmartMotionFlag: (BOOL) value {
  catSmartMotionFlag = value;
}

+(BOOL) getCatSmartMotionFlag {
  return(catSmartMotionFlag);
}

static double productionProbability = 1.0;

+(void) setProductionProbability: (double) inProductionProbability {
  productionProbability = inProductionProbability;
}

+(double) getProductionProbability {
  return(productionProbability);
}

static double catalystDecayProbability = 0.0;

+(void) setCatalystDecayProbability: (double) inCatalystDecayProbability {
  catalystDecayProbability = inCatalystDecayProbability;
}

+(double) getCatalystDecayProbability {
  return(catalystDecayProbability);
}

-(BOOL) canMoveToNeighborCode: (NeighborCode) neighborCode {
  /* For now, just implement a simplistic mechanism of
     preventing permeation through a bond along
     diagonals of the Moore neighborhood... */

  BOOL canMove;

  canMove = [super canMoveToNeighborCode:  neighborCode];
  if (canMove) {
    switch (neighborCode) {
    case SqrNorthEast:
    case SqrNorthWest:
    case SqrSouthEast:
    case SqrSouthWest:
      canMove = NO; 
      break;
    default:
    }
  }

  return (canMove);
}


-(void) countBondedNeighbors {
  NeighborCode neighborCode, initialNeighborCode;
  id neighborParticle;
  id lattice;

  singlyBondedNeighborCount = 0U;
  doublyBondedNeighborCount = 0U;
  lattice = [coord getLatticeObject];
  initialNeighborCode = [lattice getInitialNeighborCode];
  neighborCode = initialNeighborCode;
  do {
    neighborParticle = [self getAgentAtNeighborCode: neighborCode];

    if (neighborParticle == nil) 
      [InternalError raiseEvent:
	 "nil encountered at neighborCode %d of particle %d.\n",
	 neighborCode, self];
    else {
      if ([neighborParticle isMemberOf: [Link class]]) {
	switch ([neighborParticle getBondCount]) {
	case 1:
	  singlyBondedNeighborCount++;
	  break;
	case 2:
	  doublyBondedNeighborCount++;
	default:
	}
      }
    } 
    neighborCode = [lattice getNextNeighborCode: neighborCode];
  } while (neighborCode != initialNeighborCode);
}

-(void) prepareForMotion {
  if (catSmartMotionFlag)
    [self countBondedNeighbors];
}

-(double) getMobilityFactor {
  double mobilityFactor;
  unsigned oldSinglyBondedNeighborCount;
  unsigned oldDoublyBondedNeighborCount;

  mobilityFactor = catalystMobilityFactor;

  if (catSmartMotionFlag) {
    oldSinglyBondedNeighborCount = singlyBondedNeighborCount;
    oldDoublyBondedNeighborCount = doublyBondedNeighborCount;

    [self countBondedNeighbors];

    if (singlyBondedNeighborCount > oldSinglyBondedNeighborCount)
      mobilityFactor = 0.0;
    else {
      if (doublyBondedNeighborCount > oldDoublyBondedNeighborCount)
	mobilityFactor = 0.0;
    }
  }

  return mobilityFactor;
}

-doProduction {
  NeighborCode neighborCode0, neighborCode1;
  Particle *neighborParticle0, *neighborParticle1;
  Hole *hole;
  id holeCoord;
  Link *link; 
  id linkCoord;
  
  
   
  if ([bernoulli getTossWithProb: productionProbability]) {
    neighborCode0 = [[coord getLatticeObject] getRandomNeighborCode];
    neighborParticle0 = [self getAgentAtNeighborCode: neighborCode0];
    if (neighborParticle0 == nil) 
      [InternalError raiseEvent:
	 "nil encountered at neighborCode %d of particle %d.\n",
	 neighborCode0, self];
    else {
      if ([neighborParticle0 isMemberOf: [Substrate class]]) {
	if ([bernoulli getTossWithProb: 0.5])
	  neighborCode1 = 
	    [[coord getLatticeObject] getNextNeighborCode: neighborCode0]; 
	else
	  neighborCode1 = 
            [[coord getLatticeObject] getPreviousNeighborCode: neighborCode0];

	neighborParticle1 = [self getAgentAtNeighborCode: neighborCode1];
        if (neighborParticle1 == nil)
	  [InternalError raiseEvent:
	     "nil encountered at neighborCode %d of particle %d.\n",
	     neighborCode1, self];
        else {
	  if ([neighborParticle1 isMemberOf: [Substrate class]]) {
	  // OK, we got two reactants - do it!

          holeCoord = [neighborParticle0 copyCoord: [self getZone]];
	  [substrateMgr dropAgent: neighborParticle0];
	  hole = [holeMgr createAgent];
	  if (![hole attemptUnwarpToCoord: holeCoord])
	    [InternalError raiseEvent:
	      "Can't unwarp hole.\n"];
	  [holeCoord drop];

	  linkCoord = [neighborParticle1 copyCoord: [self getZone]];
	  [substrateMgr dropAgent: neighborParticle1];

	  link = [linkMgr createAgent];
	  if (![link attemptUnwarpToCoord: linkCoord])
	    [InternalError raiseEvent:
	      "Can't unwarp link.\n"];
	  [linkCoord drop];
	  }
	}
      }
    }
  }

  return self;
}

-doDecay {
  id myCoord, link;
 
  if (catalystDecayProbability != 0.0)
    if ([self isNeighboringAgentOfClass: [Catalyst class]]) 
    // Decay must be catalysed; the idea is than an
    // isolated catalyst - which may therefore be the
    // "last" one - will not decay.  This is at least a
    // *little* ad hoc ... and not even necessarily
    // desirable: it might be better if an "escaped"
    // catalyst would have a chance of "quietly"
    // disappearing...
      if ([bernoulli getTossWithProb: catalystDecayProbability]) 
      {
	myCoord = [self copyCoord: [self getZone]];
	[catalystMgr dropAgent: self];
	link = [linkMgr createAgent];
	if (![link attemptUnwarpToCoord: myCoord])
	  [InternalError raiseEvent:
	    "Failed unwarp of link %d.\n", link];
	[myCoord drop];
      } 

  return self;
}

-step {
  [self doMotion];
  [self doProduction];  
  [self doDecay];

  return self;
}

@end
